#include "stdafx.h"
#include "SelfOrganizingMap.h"

#include <iostream>

#include "image.h"

#include <assert.h>

SelfOrganizingMap::SelfOrganizingMap(int size, int mapDimensions, SomVectorFactory::DataTypes *types, int inputDimensions, LearningRateFunc *lrFunc, NbFunction *nbFunc )
{
	this->size = size;
	this->mapDimensions = mapDimensions;
	this->inputDimensions = inputDimensions;
	
	this->totalSize = (int)pow(size,mapDimensions);

	nodes = new SomNode*[int(pow(size,mapDimensions))];
	//assert(nodes!=NULL);

	// only 2d atm
	double *pos = new double[mapDimensions];
	for (int i = 0, imax=int(pow(size,mapDimensions)); i < imax; ++i )
	{
		for ( int n = 0; n < mapDimensions; ++n )
		{
			pos[n] = int(i / pow(size,n)) % size;
		//pos[1] = i/size;
		}
		
		nodes[i] = new SomNode(pos,mapDimensions);
		nodes[i]->setVector( SomVectorFactory::createVector( types, inputDimensions ) );
	}
	delete [] pos;
	
	/*
	for ( int i = 0, imax = totalSize; i < imax; ++i )
	{
		nodes[i]->print();
		cout << endl;
	}
	*/

	if ( lrFunc != NULL )
		this->lrFunc = lrFunc;
	else
		; // create std

	if ( nbFunc != NULL )
		this->nbFunc = nbFunc;
	else
		; // create std



}

int SelfOrganizingMap::getBestMatch(std::vector<double> vec)
{
	double bestDistance = nodes[0]->getSomVector()->distanceSquared(vec);
	int bestPos = 0;
	
	double distance;
	for ( int i = 1, imax = totalSize; i < imax; ++i )
	{
		distance = nodes[i]->getSomVector()->distanceSquared(vec);
		if ( distance < bestDistance )
		{
			bestDistance = distance;
			bestPos = i;
		}
	}
	return bestPos;
}

void SelfOrganizingMap::train(std::vector<std::vector<double>> trainingData, int runs)
{
	
	for ( int run = 0, rmax = runs; run < rmax; run++ )
	{
		std::cout << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b" << (double)run/rmax << "    ";
		//string s = std::to_string(run);
		//s += ".tga";
		//print2dImage(s.c_str(),128,128,trainingData);

		double nbSizeSquared = nbFunc->nbSizeSquared(run);
		double alpha = lrFunc->learningRate(run);

		for ( std::vector<vector<double>>::iterator it = trainingData.begin(); it != trainingData.end(); ++it )
		{

			double bestDistance = nodes[0]->getSomVector()->distanceSquared(*it);
			int bestPos = 0;
			double distance;
			// find best match, assume
			for ( int i = 1, imax = totalSize; i < imax; ++i )
			{
				distance = nodes[i]->getSomVector()->distanceSquared(*it);
				if ( distance < bestDistance )
				{
					bestDistance = distance;
					bestPos = i;
				}
			}

			// update neighborhood
//#pragma loop(hint_parallel(4)) // 4 threads??
//#pragma loop(ivdep) // force this shit to parallelize
			for ( int i = 0, imax = totalSize; i < imax; ++i )
			{
				if ( nodes[i]->distanceSquared( *nodes[bestPos] ) <= nbSizeSquared )
				{
					nodes[i]->getSomVector()->moveTowards( *it,  alpha );
					//nodes[i]->getSomVector()->moveTowards( *it,  alpha / (1+nodes[i]->distance( *nodes[bestPos] )) );
				}
			}

		}

	}
	cout << endl;
}

void SelfOrganizingMap::printInfo()
{
	std::cout << "SelfOrganizingMap, size: " << size << "^" << mapDimensions << std::endl;
	std::cout << "input dimensions: " << inputDimensions << std::endl;
}

void SelfOrganizingMap::print()
{
	std::cout << "SelfOrganizingMap, size: " << size << "^" << mapDimensions << std::endl;
	std::cout << "input dimensions: " << inputDimensions << std::endl;
	for ( int i = 0, imax = totalSize; i < imax; ++i )
	{
		nodes[i]->print(); std::cout << std::endl;
		//std::cout << std::endl;
	}	
}

void SelfOrganizingMap::print2dImage(const char *filename, int xSize, int ySize, std::vector<std::vector<double>> trainingData)
{
	//declare image
	int sizeYdim = 1;
	if ( this->mapDimensions == 2 )
		sizeYdim = size;

	TGAImage *img = new TGAImage( xSize, ySize );
	
	//declare a temporary color variable
	Colour c;
	
	//for ( int i = 0, imax = totalSize; i < imax; ++i )
	// clear all to white
	for ( int y = 0; y < ySize; ++y )
		for ( int x = 0; x < xSize; ++x )
		{
			c.r = 255;//(unsigned char)(nodes[i]->getSomVector()->element(0)->getValue() * 255);
			c.g = 255;//(unsigned char)(nodes[i]->getSomVector()->element(1)->getValue() * 255);
			c.b = 255;//(unsigned char)(nodes[i]->getSomVector()->element(2)->getValue() * 255);
			c.a = 255;
			//img->setPixel(c,(int)nodes[i]->getPos(0),(int)nodes[i]->getPos(1));
			img->setPixel( c, x, y );
		}


	// plot training space vectors
	for ( std::vector<vector<double>>::iterator it = trainingData.begin(); it != trainingData.end(); ++it )
	{
		c.r = 0;//(unsigned char)(nodes[i]->getSomVector()->element(0)->getValue() * 255);
		c.g = 0;//(unsigned char)(nodes[i]->getSomVector()->element(1)->getValue() * 255);
		c.b = 0;//(unsigned char)(nodes[i]->getSomVector()->element(2)->getValue() * 255);
		c.a = 255;
		
		//for ( std::vector<double>::iterator it2 = it->begin(); it2 != it->end(); ++it2 )
		//	std::cout << *it2 << ", ";
		//std::cout << std::endl;
		
		int xpos = it->at(0) * xSize;//nodes[i]->getSomVector()->element(0)->getValue()*xSize;
		int ypos = it->at(1) * ySize;//nodes[i]->getSomVector()->element(1)->getValue()*ySize;
		//int xpos = 0;
		//int ypos = 0;

		//std::cout << xpos << " , " << ypos << std::endl;

		img->setPixel( c, xpos, ypos );
		//img->setPixel(c,x,y);
	}
	


	// plot SOM input space vectors, assuming 3 nodes here (for my example images)
	for ( int i = 0, imax = totalSize; i < imax; ++i )
	{
		c.r = 255; c.g = 255; c.b = 0;

		if ( i == 0) {
			c.r = 255; c.g = 0; c.b = 0;}
		else if ( i == 1){
			c.r = 0; c.g = 255; c.b = 0;}
		else if ( i == 2){
			c.r = 0; c.g = 0; c.b = 255;}

		c.a = 255;
		
		int xpos = nodes[i]->getSomVector()->element(0)->getValue()*xSize;
		int ypos = nodes[i]->getSomVector()->element(1)->getValue()*ySize;

		img->setPixel( c, xpos, ypos );
		//img->setPixel(c,x,y);
	}
	
	//write the image to disk
	img->WriteImage(filename);
}

void SelfOrganizingMap::printColorImage(const char *filename){
	
	//declare image
	int sizeYdim = 1;
	if ( this->mapDimensions == 2 )
		sizeYdim = size;

	TGAImage *img = new TGAImage(sizeYdim,this->size);
	
	//declare a temporary color variable
	Colour c;
	
	for ( int i = 0, imax = totalSize; i < imax; ++i )
	{
		c.r = (unsigned char)(nodes[i]->getSomVector()->element(0)->getValue() * 255);
		c.g = (unsigned char)(nodes[i]->getSomVector()->element(1)->getValue() * 255);
		c.b = (unsigned char)(nodes[i]->getSomVector()->element(2)->getValue() * 255);
		c.a = 255;
		img->setPixel(c,(int)nodes[i]->getPos(0),(int)nodes[i]->getPos(1));
	}	
	
	//write the image to disk
	img->WriteImage(filename);
}

SelfOrganizingMap::~SelfOrganizingMap(void)
{
}
